/***************************************************************
 * Name:      hobbybrennen_ch_dilutorMain.cpp
 * Purpose:   Code for Application Frame
 * Author:    hefezelle (hobbybrennen.ch)
 * Created:   2018-11-28
 * Copyright: hefezelle (hobbybrennen.ch)
 * License:
 **************************************************************/

#include "wx_pch.h"
#include "hobbybrennen_ch_dilutorMain.h"
#include <wx/msgdlg.h>

#if defined(_WIN32)
    #include "windows.h"
#endif // defined

//(*InternalHeaders(hobbybrennen_ch_dilutorFrame)
#include <wx/settings.h>
#include <wx/string.h>
#include <wx/intl.h>
#include <wx/font.h>
//*)

// BEGIN NON-GUI CODE
static constexpr double TOLERANCE = 1E-12; //error tolerance used in newton iteration
static constexpr double DELTA = 1E-5; //offset used in numeric differentiation

static const double A[12] = {
 9.982012300E2,
-1.929769495E2,
 3.891238958E2,
-1.668103923E3,
 1.352215441E4,
-8.829278388E4,
 3.062874042E5,
-6.138381234E5,
 7.470172998E5,
-5.478461354E5,
 2.234460334E5,
-3.903285426E4};

static const double B[6] = {
-2.0618513E-1,
-5.2682542E-3,
 3.6130013E-5,
-3.8957702E-7,
 7.1693540E-9,
-9.9739231E-11};

static const double C[5][11] = {
{    1.693443461530087E-1,
    -1.046914743455169E1,
     7.196353469546523E1,
    -7.047478054272792E2,
     3.924090430035045E3,
    -1.210164659068747E4,
     2.248646550400788E4,
    -2.605562982188164E4,
     1.852373922069467E4,
    -7.420201433430137E3,
     1.285617841998974E3
},
{   -1.193013005057010E-2,
     2.517399633803461E-1,
    -2.170575700536993E0,
     1.353034988843029E1,
    -5.029988758547014E1,
     1.096355666577570E2,
    -1.422753946421155E2,
     1.080435942856230E2,
    -4.414153236817392E1,
     7.442971530188783E0
},
{   -6.802995733503803E-4,
     1.876837790289664E-2,
    -2.002561813734156E-1,
     1.022992966719220E0,
    -2.895696483903638E0,
     4.810060584300675E0,
    -4.672147440794683E0,
     2.458043105903461E0,
    -5.411227621436812E-1
},
{    4.075376675622027E-6,
    -8.763058573471110E-6,
     6.515031360099368E-6,
    -1.515784836987210E-6
},
{   -2.788074354782409E-8,
     1.345612883493354E-8
}};

double mixture_density(double abw, double temperature)
{
    //this implements
    //https://www.oiml.org/en/files/pdf_r/r022-e75.pdf/view
    double result = 0.0;
    double temp = temperature - 20;

    //deal with A constants
    double abw_power = 1.0;
    for (int k=1; k<=12; ++k)
    {
        result += A[k-1]*abw_power;
        abw_power *= abw;
    }

    //deal with B constants
    double temp_power = temp;
    for (int k=1; k<=6; ++k)
    {
        result += B[k-1]*temp_power;
        temp_power *= temp;
    }

    //deal with C constants
    temp_power = temp;
    for (int i=1; i<=5; ++i)
    {
        abw_power = abw;
        for (int k=1; k<=11; ++k)
        {
            result += C[i-1][k-1]*abw_power*temp_power;
            abw_power *= abw;
        }
        temp_power *= temp;
    }

    return result/1000;
}

double water_density(double temperature)
{
    return  mixture_density(0, temperature);
}

double ethanol_density(double temperature)
{
    //taken from
    //https://en.wikipedia.org/wiki/Ethanol_(data_page)#Density_of_ethanol_at_various_temperatures
    return 0.8063372 - temperature*8.461834E-4;
}

double abv_from_abw(double abw, double temperature)
{
    //for a unit volume, density=mass.
    //abw is the fraction of mass that belongs to ethanol.
    //hence, abw*mixture_density equals the mass of alcohol in a unit volume of mixture.
    //we divide that by alcohol's density, to get the volume of pure alc.
    //in a unit volume of mixture, the volume of pure alc is the abv.
    return abw*mixture_density(abw, temperature)/ethanol_density(temperature);
}

double abw_from_abv(double abv, double temperature)
{
    //this uses newton iteration to invert the abv_from_abw function
    double approx_abw = abv;
    double error = 1E6;

    double current_value = abv_from_abw(approx_abw, temperature);
    while (error>TOLERANCE)
    {
        approx_abw -= (current_value-abv)/(abv_from_abw(approx_abw+DELTA, temperature)-current_value)*DELTA;
        current_value = abv_from_abw(approx_abw, temperature);
        error = abs(current_value-abv);
    }

    return approx_abw;
}

double dilute_unit_volume(double source_abv, double target_abv, double temperature)
{
    double x = abw_from_abv(source_abv, temperature);
    double y = abw_from_abv(target_abv, temperature);

    //return mixture_density(x, temperature)*(x/y - 1)/water_density(temperature);
    return source_abv * ethanol_density(temperature) * (1/y - 1/x) / water_density(temperature);
    //return ( source_abv*ethanol_density(temperature)/y - mixture_density(x, temperature) )/water_density(temperature);
}
// END NON-GUI CODE

//helper functions
enum wxbuildinfoformat {
    short_f, long_f };

wxString wxbuildinfo(wxbuildinfoformat format)
{
    wxString wxbuild(wxVERSION_STRING);

    if (format == long_f )
    {
#if defined(__WXMSW__)
        wxbuild << _T("-Windows");
#elif defined(__UNIX__)
        wxbuild << _T("-Linux");
#endif

#if wxUSE_UNICODE
        wxbuild << _T("-Unicode build");
#else
        wxbuild << _T("-ANSI build");
#endif // wxUSE_UNICODE
    }

    return wxbuild;
}

//(*IdInit(hobbybrennen_ch_dilutorFrame)
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT1 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_SLIDER1 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_TEXTCTRL1 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT5 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_SLIDER4 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_TEXTCTRL4 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT2 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_SLIDER2 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_TEXTCTRL2 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT3 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_SLIDER3 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_TEXTCTRL3 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICLINE1 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT4 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT6 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATICTEXT7 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_PANEL1 = wxNewId();
const long hobbybrennen_ch_dilutorFrame::idMenuQuit = wxNewId();
const long hobbybrennen_ch_dilutorFrame::idMenuAbout = wxNewId();
const long hobbybrennen_ch_dilutorFrame::ID_STATUSBAR1 = wxNewId();
//*)

BEGIN_EVENT_TABLE(hobbybrennen_ch_dilutorFrame,wxFrame)
    //(*EventTable(hobbybrennen_ch_dilutorFrame)
    //*)
END_EVENT_TABLE()

hobbybrennen_ch_dilutorFrame::hobbybrennen_ch_dilutorFrame(wxWindow* parent,wxWindowID id)
{
    //(*Initialize(hobbybrennen_ch_dilutorFrame)
    wxMenuItem* MenuItem2;
    wxMenuItem* MenuItem1;
    wxFlexGridSizer* FlexGridSizer1;
    wxMenu* Menu1;
    wxBoxSizer* BoxSizer1;
    wxMenuBar* MenuBar1;
    wxMenu* Menu2;

    Create(parent, id, _("hobbybrennen.ch Verduennungsrechner"), wxDefaultPosition, wxDefaultSize, wxDEFAULT_FRAME_STYLE, _T("id"));
    SetClientSize(wxSize(600,400));
    Panel1 = new wxPanel(this, ID_PANEL1, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL, _T("ID_PANEL1"));
    BoxSizer1 = new wxBoxSizer(wxVERTICAL);
    FlexGridSizer2 = new wxFlexGridSizer(5, 1, 0, 0);
    FlexGridSizer2->AddGrowableCol(0);
    FlexGridSizer2->AddGrowableRow(4);
    FlexGridSizer1 = new wxFlexGridSizer(4, 3, 0, 0);
    FlexGridSizer1->AddGrowableCol(1);
    StaticText1 = new wxStaticText(Panel1, ID_STATICTEXT1, _("Temperatur in C"), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT1"));
    FlexGridSizer1->Add(StaticText1, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    SliderTemp = new wxSlider(Panel1, ID_SLIDER1, 2200, -1000, 4000, wxDefaultPosition, wxDefaultSize, 0, wxDefaultValidator, _T("ID_SLIDER1"));
    FlexGridSizer1->Add(SliderTemp, 1, wxALL|wxEXPAND, 5);
    TextTemp = new wxTextCtrl(Panel1, ID_TEXTCTRL1, _("22.25"), wxDefaultPosition, wxSize(64,36), wxTE_PROCESS_ENTER, wxDefaultValidator, _T("ID_TEXTCTRL1"));
    TextTemp->SetMaxLength(5);
    FlexGridSizer1->Add(TextTemp, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    StaticText4 = new wxStaticText(Panel1, ID_STATICTEXT5, _("Ausgangs-Vol in L"), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT5"));
    FlexGridSizer1->Add(StaticText4, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    SliderSourceVol = new wxSlider(Panel1, ID_SLIDER4, 100, 0, 2000, wxDefaultPosition, wxDefaultSize, 0, wxDefaultValidator, _T("ID_SLIDER4"));
    FlexGridSizer1->Add(SliderSourceVol, 1, wxALL|wxEXPAND, 5);
    TextSourceVol = new wxTextCtrl(Panel1, ID_TEXTCTRL4, _("1.00"), wxDefaultPosition, wxSize(64,36), wxTE_PROCESS_ENTER, wxDefaultValidator, _T("ID_TEXTCTRL4"));
    FlexGridSizer1->Add(TextSourceVol, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    StaticText2 = new wxStaticText(Panel1, ID_STATICTEXT2, _("Ausgangs-Vol%"), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT2"));
    FlexGridSizer1->Add(StaticText2, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    SliderSource = new wxSlider(Panel1, ID_SLIDER2, 4000, 0, 10000, wxDefaultPosition, wxDefaultSize, 0, wxDefaultValidator, _T("ID_SLIDER2"));
    FlexGridSizer1->Add(SliderSource, 1, wxALL|wxEXPAND, 5);
    TextSource = new wxTextCtrl(Panel1, ID_TEXTCTRL2, _("40.00"), wxDefaultPosition, wxSize(64,36), wxTE_PROCESS_ENTER, wxDefaultValidator, _T("ID_TEXTCTRL2"));
    TextSource->SetMaxLength(5);
    FlexGridSizer1->Add(TextSource, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    StaticText3 = new wxStaticText(Panel1, ID_STATICTEXT3, _("Ziel-Vol%"), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT3"));
    FlexGridSizer1->Add(StaticText3, 1, wxALL|wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL, 5);
    SliderTarget = new wxSlider(Panel1, ID_SLIDER3, 3200, 0, 10000, wxDefaultPosition, wxDefaultSize, 0, wxDefaultValidator, _T("ID_SLIDER3"));
    FlexGridSizer1->Add(SliderTarget, 1, wxALL|wxEXPAND, 5);
    TextTarget = new wxTextCtrl(Panel1, ID_TEXTCTRL3, _("32.00"), wxDefaultPosition, wxSize(64,36), wxTE_PROCESS_ENTER, wxDefaultValidator, _T("ID_TEXTCTRL3"));
    TextTarget->SetMaxLength(5);
    FlexGridSizer1->Add(TextTarget, 1, wxALL|wxALIGN_RIGHT, 5);
    FlexGridSizer2->Add(FlexGridSizer1, 1, wxALL|wxEXPAND, 5);
    StaticLine1 = new wxStaticLine(Panel1, ID_STATICLINE1, wxDefaultPosition, wxSize(10,-1), wxLI_HORIZONTAL, _T("ID_STATICLINE1"));
    FlexGridSizer2->Add(StaticLine1, 1, wxALL|wxEXPAND, 5);
    result_text1 = new wxStaticText(Panel1, ID_STATICTEXT4, _("Um 1 Liter bei 22.25 Grad Celsius von 40%vol\nauf 32%vol zu verduennen, muss man"), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT4"));
    FlexGridSizer2->Add(result_text1, 1, wxALL|wxALIGN_TOP|wxALIGN_CENTER_HORIZONTAL, 5);
    result_ml = new wxStaticText(Panel1, ID_STATICTEXT6, _("252"), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT6"));
    wxFont result_mlFont = wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT);
    if ( !result_mlFont.Ok() ) result_mlFont = wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT);
    result_mlFont.SetPointSize(20);
    result_mlFont.SetStyle(wxFONTSTYLE_NORMAL);
    result_mlFont.SetWeight(wxFONTWEIGHT_BOLD);
    result_mlFont.SetUnderlined(false);
    result_mlFont.SetFamily(wxFONTFAMILY_SWISS);
    result_mlFont.SetFaceName(_T("Sans"));
    result_ml->SetFont(result_mlFont);
    FlexGridSizer2->Add(result_ml, 1, wxALL|wxALIGN_TOP|wxALIGN_CENTER_HORIZONTAL, 5);
    result_text2 = new wxStaticText(Panel1, ID_STATICTEXT7, _("Milliliter Wasser zugeben."), wxDefaultPosition, wxDefaultSize, 0, _T("ID_STATICTEXT7"));
    FlexGridSizer2->Add(result_text2, 1, wxALL|wxALIGN_TOP|wxALIGN_CENTER_HORIZONTAL, 5);
    BoxSizer1->Add(FlexGridSizer2, 1, wxALL|wxEXPAND, 5);
    Panel1->SetSizer(BoxSizer1);
    BoxSizer1->Fit(Panel1);
    BoxSizer1->SetSizeHints(Panel1);
    MenuBar1 = new wxMenuBar();
    Menu1 = new wxMenu();
    MenuItem1 = new wxMenuItem(Menu1, idMenuQuit, _("Quit\tAlt-F4"), _("Quit the application"), wxITEM_NORMAL);
    Menu1->Append(MenuItem1);
    MenuBar1->Append(Menu1, _("&File"));
    Menu2 = new wxMenu();
    MenuItem2 = new wxMenuItem(Menu2, idMenuAbout, _("About\tF1"), _("Show info about this application"), wxITEM_NORMAL);
    Menu2->Append(MenuItem2);
    MenuBar1->Append(Menu2, _("Help"));
    SetMenuBar(MenuBar1);
    StatusBar1 = new wxStatusBar(this, ID_STATUSBAR1, 0, _T("ID_STATUSBAR1"));
    int __wxStatusBarWidths_1[1] = { -1 };
    int __wxStatusBarStyles_1[1] = { wxSB_NORMAL };
    StatusBar1->SetFieldsCount(1,__wxStatusBarWidths_1);
    StatusBar1->SetStatusStyles(1,__wxStatusBarStyles_1);
    SetStatusBar(StatusBar1);

    Connect(ID_SLIDER1,wxEVT_SCROLL_THUMBTRACK,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnSliderTempCmdScroll);
    Connect(ID_TEXTCTRL1,wxEVT_COMMAND_TEXT_ENTER,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnTextTempTextEnter);
    Connect(ID_SLIDER4,wxEVT_SCROLL_THUMBTRACK,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnSliderSourceVolCmdScroll);
    Connect(ID_TEXTCTRL4,wxEVT_COMMAND_TEXT_ENTER,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnTextSourceVolTextEnter);
    Connect(ID_SLIDER2,wxEVT_SCROLL_THUMBTRACK,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnSliderSourceCmdScroll);
    Connect(ID_TEXTCTRL2,wxEVT_COMMAND_TEXT_ENTER,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnTextSourceTextEnter);
    Connect(ID_SLIDER3,wxEVT_SCROLL_THUMBTRACK,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnSliderTargetCmdScroll);
    Connect(ID_TEXTCTRL3,wxEVT_COMMAND_TEXT_ENTER,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnTextTargetTextEnter);
    Connect(idMenuQuit,wxEVT_COMMAND_MENU_SELECTED,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnQuit);
    Connect(idMenuAbout,wxEVT_COMMAND_MENU_SELECTED,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnAbout);
    Connect(wxID_ANY,wxEVT_CLOSE_WINDOW,(wxObjectEventFunction)&hobbybrennen_ch_dilutorFrame::OnClose);
    //*)
}

hobbybrennen_ch_dilutorFrame::~hobbybrennen_ch_dilutorFrame()
{
    //(*Destroy(hobbybrennen_ch_dilutorFrame)
    //*)
}

void hobbybrennen_ch_dilutorFrame::OnQuit(wxCommandEvent& event)
{
    Close();
}

void hobbybrennen_ch_dilutorFrame::OnAbout(wxCommandEvent& event)
{
    wxString msg;
    wxString build = wxbuildinfo(long_f);
    msg << "Version 1.0 (2018-12-03)\nImplementierung: hefezelle\n"
    << "Hilfe: azeotrop, Huegelwilli\n\n"
    << build.ToStdString();
    wxMessageBox(msg, _("Info"));
}

void hobbybrennen_ch_dilutorFrame::OnClose(wxCloseEvent& event)
{
    Close();
}

void hobbybrennen_ch_dilutorFrame::update_text()
{
    wxString text1;
    wxString text2;
    auto temp = atof(hobbybrennen_ch_dilutorFrame::TextTemp->GetValue());
    auto src_vol = atof(hobbybrennen_ch_dilutorFrame::TextSourceVol->GetValue());
    auto src = atof(hobbybrennen_ch_dilutorFrame::TextSource->GetValue());
    auto targ = atof(hobbybrennen_ch_dilutorFrame::TextTarget->GetValue());
    double sol = src_vol*dilute_unit_volume(src/100, targ/100, temp);

    text1 << "Um " << src_vol << " Liter bei " << temp << " Grad Celsius von "
    << src << "%vol\nauf " << targ << "%vol zu verduennen, muss man";
    text2 << round(sol*10000)/10.0;
    hobbybrennen_ch_dilutorFrame::result_text1->SetLabel(text1);
    hobbybrennen_ch_dilutorFrame::result_ml->SetLabel(text2);
    hobbybrennen_ch_dilutorFrame::FlexGridSizer2->Layout();
}

void hobbybrennen_ch_dilutorFrame::OnSliderTempCmdScroll(wxScrollEvent& event)
{
    wxString value;
    value << hobbybrennen_ch_dilutorFrame::SliderTemp->GetValue()/100.0;
    hobbybrennen_ch_dilutorFrame::TextTemp->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnSliderSourceVolCmdScroll(wxScrollEvent& event)
{
    wxString value;
    value << hobbybrennen_ch_dilutorFrame::SliderSourceVol->GetValue()/100.0;
    hobbybrennen_ch_dilutorFrame::TextSourceVol->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnSliderSourceCmdScroll(wxScrollEvent& event)
{
    wxString value;
    value << hobbybrennen_ch_dilutorFrame::SliderSource->GetValue()/100.0;
    hobbybrennen_ch_dilutorFrame::TextSource->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnSliderTargetCmdScroll(wxScrollEvent& event)
{
    wxString value;
    value << hobbybrennen_ch_dilutorFrame::SliderTarget->GetValue()/100.0;
    hobbybrennen_ch_dilutorFrame::TextTarget->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnTextTempTextEnter(wxCommandEvent& event)
{
    int value = round(atof(hobbybrennen_ch_dilutorFrame::TextTemp->GetValue())*100);
    hobbybrennen_ch_dilutorFrame::SliderTemp->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnTextSourceVolTextEnter(wxCommandEvent& event)
{
    int value = round(atof(hobbybrennen_ch_dilutorFrame::TextSourceVol->GetValue())*100);
    hobbybrennen_ch_dilutorFrame::SliderSourceVol->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnTextSourceTextEnter(wxCommandEvent& event)
{
    int value = round(atof(hobbybrennen_ch_dilutorFrame::TextSource->GetValue())*100);
    hobbybrennen_ch_dilutorFrame::SliderSource->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}

void hobbybrennen_ch_dilutorFrame::OnTextTargetTextEnter(wxCommandEvent& event)
{
    int value = round(atof(hobbybrennen_ch_dilutorFrame::TextTarget->GetValue())*100);
    hobbybrennen_ch_dilutorFrame::SliderTarget->SetValue(value);
    hobbybrennen_ch_dilutorFrame::update_text();
}
